# -----------------------------------------------------------------------------
# Show Mardigras constraint violations using Chimera
#
import string
import Tkinter

import chimera
import chimera.selection

import atoms
import mardigras
import noesy
import pdb
import sparky
import sputil
import tkutil

# -----------------------------------------------------------------------------
#
class noeshow_dialog(tkutil.Dialog):

  def __init__(self):

    self.pdb_path = None
    self.model = None
    self.constraints = None
    self.recenter_on_peak = 1
    self.chimera_atom_to_num_atom = {}
    self.num_atom_to_chimera_atom = {}
    self.chimera_atom_to_resonance = {}
    self.num_atom_pair_to_pseudobond = {}
    
    self.peak_selected_callback = self.peak_selected_cb
    sparky.add_selection_callback(self.peak_selected_callback)
    
    chimera.triggers.addHandler('selectionChanged',
				self.chimera_selection_cb, None)
    chimera.viewer.highlight = chimera.Viewer_Outline

    tkutil.Dialog.__init__(self, 'View Constraints')

    self.pdb_file = tkutil.file_field(self.top, 'PDB File: ', 'pdb')
    self.pdb_file.frame.pack(side = 'top', anchor = 'w')

    self.mardi_file = tkutil.file_field(self.top,
					'Mardigras Constraint File: ',
					'mardigras')
    self.mardi_file.frame.pack(side = 'top', anchor = 'w')

    cb = tkutil.checkbuttons(self.top, 'top', '',
                             'Show satisfied constraints?',
                             'Show violated constraints?')
    cb.frame.pack(side = 'top', anchor = 'w')
    self.satisfied, self.violated = cb.variables
    self.satisfied.set(0)

    colors = ('white', 'black', 'red', 'yellow',
              'green', 'cyan', 'blue', 'magenta')
    c = tkutil.option_menu(self.top, 'Color assigned atoms: ',
                           colors, 'yellow')
    c.frame.pack(side = 'top', anchor = 'w')
    self.assigned_atom_color = c
    
    self.status_line = Tkinter.Label(self.top, anchor = 'nw')
    self.status_line.pack(side = 'top', anchor = 'w')

    br = tkutil.button_row(self.top,
			   ('Show', self.show_cb),
			   ('Close', self.close_cb),
                           ('Help', sputil.help_cb('Chimera')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

  # -------------------------------------------------------------------------
  #
  def show_cb(self):

    self.show()

  # -------------------------------------------------------------------------
  #
  def show(self):

    pdb_path = self.pdb_file.get()
    assigned_color = self.assigned_atom_color.get()
    
    self.chimera_open_model(pdb_path)
    self.color_assigned_atoms(assigned_color)

    mardi_path = self.mardi_file.get()
    self.show_constraints(mardi_path, pdb_path)

  # -------------------------------------------------------------------------
  #
  def chimera_open_model(self, pdb_path):
      
    if pdb_path == self.pdb_path:
      return

    if self.model:
      self.openModels.close([self.model])
      self.model = None
      # I suspect closing the model also eliminates all constraint pseudobonds
      self.num_atom_pair_to_pseudobond = {}

    self.pdb_path = pdb_path
    models = chimera.openModels.open(pdb_path, type='PDB')
    self.model = models[0]    # assume just one molecule in the PDB file
    self.model.color = chimera.Color_lookup('gray')

    trans = pdb.model(pdb_path).atom_translations
    self.make_atom_tables(self.model.atoms(), trans)

    if self.constraints == None:
      m = chimera.PseudoBondMgr_mgr()
      self.constraints = m.newPseudoBondGroup('NOESY constraints')
      del m

    chimera.openModels._add([self.constraints], sameAs = self.model.id)

  # -------------------------------------------------------------------------
  #
  def make_atom_tables(self, chimera_atoms, translations):

    ca2na = {}
    na2ca = {}
    for ca in chimera_atoms:
      na = self.chimera_num_atom(ca, translations)
      ca2na[ca] = na
      na2ca[na] = ca
    self.chimera_atom_to_num_atom = ca2na
    self.num_atom_to_chimera_atom = na2ca

    ca2r = {}
    condition = sputil.selected_condition()
    if condition:
      for r in condition.resonance_list():
	num_atom = self.resonance_num_atom(r)
	if na2ca.has_key(num_atom):
	  ca = na2ca[num_atom]
	  ca2r[ca] = r
    self.chimera_atom_to_resonance = ca2r
  
  # -------------------------------------------------------------------------
  #
  def color_assigned_atoms(self, colorname):

    color = chimera.Color_lookup(colorname)
    if color == None:
      return
    
    for ca, r in self.chimera_atom_to_resonance.items():
      if r.peak_count > 0:
	ca.color = color

  # -------------------------------------------------------------------------
  # Example :23@CA
  #
  def chimera_atom_name(self, chimera_atom):

    return chimera_atom.oslIdent(start=chimera.SelResidue)

  # -------------------------------------------------------------------------
  #
  def chimera_num_atom(self, chimera_atom, translations):
    
    cname = self.chimera_atom_name(chimera_atom)
    num_string, atom = string.split(cname[1:], '@')
    num = string.atoi(num_string)
    return standard_num_atom(num, atom, translations)

  # -------------------------------------------------------------------------
  #
  def resonance_num_atom(self, r):

    trans = atoms.molecule_atom_name_translations(r.condition.molecule)
    return standard_num_atom(r.group.number, r.atom.name, trans)
  
  # -------------------------------------------------------------------------
  # Show constraints colored by violation
  #
  def show_constraints(self, mardi_path, pdb_path):

    # unshow previous constraints

    if not mardi_path or not pdb_path:
      return

    show_satisfied = self.satisfied.get()
    show_violated = self.violated.get()
    if not (show_satisfied or show_violated):
      return
    
    pdb_model = pdb.model(pdb_path)
    mardi = mardigras.bounds(mardi_path)
    for (num_atom_1, num_atom_2), (rmin, rmax) in mardi.bounds_table.items():
      num_atom_1 = translate_atom_name(num_atom_1, mardi, pdb_model)
      num_atom_2 = translate_atom_name(num_atom_2, mardi, pdb_model)
      d = pdb_model.distance(num_atom_1, num_atom_2)
      if d != None:
        if ((show_satisfied and d >= rmin and d <= rmax) or
            (show_violated and (d < rmin or d > rmax))):
          if d < .8 * rmin:	color = 'blue'
          elif d < rmin:	color = 'cyan'
          elif d <= rmax:	color = 'green'
          elif d < 1.2 * rmax:	color = 'magenta'
          else:			color = 'red'
	  self.show_constraint(num_atom_1, num_atom_2, color)

  # -------------------------------------------------------------------------
  #
  def show_constraint(self, num_atom_1, num_atom_2, color):
    
    if (self.num_atom_to_chimera_atom.has_key(num_atom_1) and
	self.num_atom_to_chimera_atom.has_key(num_atom_2)):
      a1 = self.num_atom_to_chimera_atom[num_atom_1]
      a2 = self.num_atom_to_chimera_atom[num_atom_2]
      b = self.constraints.newPseudoBond(a1, a2)
      b.color = chimera.Color_lookup(color)
      b.display = 1
      b.halfbond = 0
      atom_pair = self.num_atom_pair(num_atom_1, num_atom_2)
      self.num_atom_pair_to_pseudobond[atom_pair] = b
      return b
    return None

  # -------------------------------------------------------------------------
  #
  def find_constraint(self, num_atom_1, num_atom_2):
    
    atom_pair = self.num_atom_pair(num_atom_1, num_atom_2)
    if self.num_atom_pair_to_pseudobond.has_key(atom_pair):
      return self.num_atom_pair_to_pseudobond[atom_pair]
    return None
    
  # ---------------------------------------------------------------------------
  #
  def num_atom_pair(self, num_atom_1, num_atom_2):

    pair = [num_atom_1, num_atom_2]
    pair.sort()
    return tuple(pair)
    
  # ---------------------------------------------------------------------------
  #
  def peak_selected_cb(self):

    if self.is_window_destroyed():
      if self.peak_selected_callback:
        sparky.remove_selection_callback(self.peak_selected_callback)
      return

    peak = sputil.selected_peak()
    if peak:
      self.show_peak_constraint(peak)

  # -------------------------------------------------------------------------
  #
  def show_peak_constraint(self, peak):

    rlist = peak.resonances()
    if len(rlist) == 2:
      r1, r2 = rlist
      if r1 and r2:
        self.status_line['text'] = 'Selected ' + peak.assignment
        num_atom_1 = self.resonance_num_atom(r1)
        num_atom_2 = self.resonance_num_atom(r2)
        b = self.find_constraint(num_atom_1, num_atom_2)
        if b == None:
          b = self.show_constraint(num_atom_1, num_atom_2, 'blue')
        if b == None:
          message ='Missing PDB atoms for ' + peak.assignment
          self.status_line['text'] = message
        else:
          self.recenter_on_peak = 0
          chimera.selection.setCurrent(b)     # Set Chimera selection
          self.recenter_on_peak = 1
          
  # -------------------------------------------------------------------------
  #
  def chimera_selection_cb(self, trigger, closure, selection):

    if not self.recenter_on_peak:
      return

    edges = selection.itemize(edge=1)[chimera.EDGE]
    edges = filter(lambda e: isinstance(e, chimera.PseudoBondType), edges)
    edges = filter(lambda e, c=self.constraints: e.pseudoBondGroup == c, edges)
    if len(edges) == 1:
      edge = edges[0]
      atom1, atom2 = edge.atoms()

      s = sputil.selected_spectrum()
      if s == None or not noesy.is_noesy(s):
	return

      r1 = self.chimera_atom_resonance(atom1)
      r2 = self.chimera_atom_resonance(atom2)
      if r1 and r2:
	sputil.show_assignment((r1, r2), s)
        sputil.select_assignment((r1, r2), s)
	message = 'Showing %s-%s in spectrum %s' % (r1.name, r2.name, s.name)
	self.status_line['text'] = message
      else:
	message = 'No resonance'
	if not r1: message = message + ' ' + self.chimera_atom_name(atom1)
	if not r2: message = message + ' ' + self.chimera_atom_name(atom2)
	message = message + ' in spectrum ' + s.name
	self.status_line['text'] = message

  # -------------------------------------------------------------------------
  #
  def chimera_atom_resonance(self, chimera_atom):

    if self.chimera_atom_to_resonance.has_key(chimera_atom):
      return self.chimera_atom_to_resonance[chimera_atom]
    return None

# -----------------------------------------------------------------------------
# Translate a mardigras (residue #, atom name) to pdb file format
#
def translate_atom_name(num_atom, mardi, model):

  g, a = atoms.translate_group_atom(None, num_atom[1],
                                    mardi.atom_translations,
                                    model.atom_translations)
  return (num_atom[0], a)

# -----------------------------------------------------------------------------
# Apply atom name translations.
#
def standard_num_atom(num, atom, translations):

  g, a = translations.to_standard_name(None, atom)
  return (num, a)

# -----------------------------------------------------------------------------
#
def show_chimera_dialog():
  tkutil.the_dialog(noeshow_dialog).show_window(1)
