# -----------------------------------------------------------------------------
# Use structure atom distances to help fixup assignments.  For example,
# find possible assignments for NOE peaks where atoms are within 5 angstroms
# and resonances are within .03 ppm.  Or find all proton pairs within 5
# angstroms for which there is no assigned peak.
#
import string
import Tkinter
import types

import atoms
import noesy
import pdb
import pyutil
import sparky
import sputil
import tkutil

# -----------------------------------------------------------------------------
#
class assignment_distance_dialog(tkutil.Dialog, tkutil.Stoppable):

  # ---------------------------------------------------------------------------
  #
  def __init__(self):

    self.show_selected_cb = None
    self.make_dialog()

  # ---------------------------------------------------------------------------
  #
  def make_dialog(self):

    tkutil.Dialog.__init__(self, 'Assignment Distances')

    self.top.bind('<Destroy>', self.window_destroyed_cb)

    self.pdb_path = tkutil.file_field(self.top, 'PDB file: ', 'pdb')
    self.pdb_path.frame.pack(side = 'top', anchor = 'w')

    self.spectrum_choice = sputil.spectrum_menu(self.top, 'Spectrum: ')
    self.spectrum_choice.frame.pack(side = 'top', anchor = 'w')
    self.spectrum_widget = self.spectrum_choice.frame
    self.spectrum_choice.add_callback(self.spectrum_cb)

    er = tkutil.entry_row(self.top, 'PPM tolerance: ',
		       ('w1', '.01', 4), ('w2', '.01', 4), ('w3', '.01', 4))
    self.ppm_range = er.variables
    self.w3_range_widget = er.entries[2].frame
    er.frame.pack(side = 'top', anchor = 'w')

    self.line_format = '%15s %5s %7s %s'

    self.spectrum_cb(self.spectrum_choice.get())

    e = tkutil.entry_field(self.top, 'Max atom distance: ', '5.5', 5)
    self.max_dist = e.variable
    e.frame.pack(side = 'top', anchor = 'w')

    choices = ('Peaks with far atoms',
	       'Multiple assignments',
	       'Close atoms with no peak',
	       'Unassigned peak assignments',
	       'Unassignable peaks',
	       'Selected peak assignments')
    self.show_type = tkutil.option_menu(self.top, 'Show: ',
                                        choices, choices[0])
    self.show_type.frame.pack(side = 'top', anchor = 'w')

    pl = tkutil.scrolling_list(self.top, '', 5)
    pl.frame.pack(side = 'top', fill = 'both', expand = 1)
    pl.listbox.bind('<ButtonRelease-1>', self.select_peak_or_assignment_cb)
    pl.listbox.bind('<Double-ButtonRelease-1>',
                    self.goto_peak_or_assignment_cb)
    self.peak_list = pl

    progress_label = Tkinter.Label(self.top, anchor = 'nw')
    progress_label.pack(side = 'top', anchor = 'nw')

    br = tkutil.button_row(self.top,
			   ('Update', self.update_cb),
			   ('Save', self.peak_list.save_cb),
			   ('Stop', self.stop_cb),
                           ('Close', self.close_cb),
                           ('Help', sputil.help_cb('AssignmentDistance')),
			   )
    br.frame.pack(side = 'top', anchor = 'w')

    tkutil.Stoppable.__init__(self, progress_label, br.buttons[2])

    self.settings = self.get_settings()

  # ---------------------------------------------------------------------------
  #
  def window_destroyed_cb(self, event):
    if self.show_selected_cb != None:
      sparky.remove_selection_callback(self.show_selected_cb)
      self.show_selected_cb = None

  # ---------------------------------------------------------------------------
  #
  def spectrum_cb(self, name):
    spectrum = sputil.name_to_spectrum(name)
    if spectrum:
      if spectrum.dimension == 3:
        self.w3_range_widget.pack(side = 'left', anchor = 'w')
        self.line_format = '%20s %5s %7s %s'
      else:
        self.w3_range_widget.pack_forget()
        self.line_format = '%15s %5s %7s %s'

  # ---------------------------------------------------------------------------
  #
  def update_cb(self):

    s = self.get_settings()

    self.settings = s
    self.model = pdb.model(s.pdb_path)
    self.peak_list.spectrum = s.spectrum

    if s.show_type == 'Peaks with far atoms':
      self.stoppable_call(self.show_peaks,
			  self.show_far_atom_peak,
			  'excessive distance peaks')
    elif s.show_type == 'Multiple assignments':
      self.stoppable_call(self.show_peaks,
			  self.show_multiple_assignments,
			  'multiply assignable peaks')
    elif s.show_type == 'Close atoms with no peak':
      self.stoppable_call(self.show_missing_assignments)
    elif s.show_type == 'Unassigned peak assignments':
      self.stoppable_call(self.show_peaks,
			  self.show_assignable_unassigned_peak,
			  'assignable unassigned peaks')
    elif s.show_type == 'Unassignable peaks':
      self.stoppable_call(self.show_peaks,
			  self.show_unassignable_unassigned_peak,
			  'unassignable unassigned peaks')
    elif s.show_type == 'Selected peak assignments':
      if self.show_selected_cb == None:
	self.show_selected_cb = self.show_selected_peak_assignments
	sparky.add_selection_callback(self.show_selected_cb)
      self.show_selected_cb()

    if (self.show_selected_cb != None and
	s.show_type != 'Selected peak assignments'):
      sparky.remove_selection_callback(self.show_selected_cb)
      self.show_selected_cb = None

  # ---------------------------------------------------------------------------
  #
  def get_settings(self):

    settings = pyutil.generic_class()
    settings.spectrum = sputil.name_to_spectrum(self.spectrum_choice.get())
    settings.noesy_hsqc = noesy.is_noesy_hsqc(settings.spectrum)
    settings.pdb_path = self.pdb_path.get()
    settings.max_distance = string.atof(self.max_dist.get())
    if settings.spectrum:
      dim = settings.spectrum.dimension
      settings.ppm_range = tkutil.float_variable_values(self.ppm_range)[:dim]
    else:
      settings.ppm_range = None
    settings.show_type = self.show_type.get()
    return settings

  # ---------------------------------------------------------------------------
  #
  def select_peak_or_assignment_cb(self, event):

    pora = self.peak_list.event_line_data(event)
    if type(pora) == types.InstanceType:
      sputil.select_peak(pora)
    elif type(pora) == types.TupleType or type(pora) == types.ListType:
      sputil.select_peak(self.settings.spectrum.find_peak(pora))
      
  # ---------------------------------------------------------------------------
  #
  def goto_peak_or_assignment_cb(self, event):

    pora = self.peak_list.event_line_data(event)
    if type(pora) == types.InstanceType:
      sputil.show_peak(pora)
    elif type(pora) == types.TupleType or type(pora) == types.ListType:
      sputil.show_assignment(pora, self.settings.spectrum)

  # ---------------------------------------------------------------------------
  #
  def show_peaks(self, show_peak, heading, peaks = None):

    self.peak_list.clear()

    if peaks == None:
      if self.settings.spectrum:
        peaks = self.settings.spectrum.peak_list()
      else:
        peaks = []

    shown = 0
    self.stoppable_loop('peaks', 50)
    for peak in peaks:
      self.check_for_stop()
      if show_peak(peak):
        shown = shown + 1

    self.peak_list.heading['text'] = ('%d %s\n' % (shown, heading)
				      + '  ' + self.line_format %
				      ('assignment', 'dist', 'S/N',
				      '      offset        '))

  # ---------------------------------------------------------------------------
  #
  def show_far_atom_peak(self, peak):
    dist = self.assignment_distance(peak.resonances())
    return (dist != None and
	    dist > self.settings.max_distance and
	    self.show_alternative_assignments(peak))

  # ---------------------------------------------------------------------------
  #
  def show_multiple_assignments(self, peak):
    return self.show_alternative_assignments(peak, 2)

  # ---------------------------------------------------------------------------
  #
  def show_assignable_unassigned_peak(self, peak):
    return not peak.is_assigned and self.show_alternative_assignments(peak, 1)

  # ---------------------------------------------------------------------------
  #
  def show_unassignable_unassigned_peak(self, peak):
    if peak.is_assigned:
      return 0
    alist = self.possible_assignments(peak)
    if len(alist) > 0:
      return 0
    self.peak_list.append(self.peak_format(peak), peak)
    return 1

  # ---------------------------------------------------------------------------
  #
  def show_missing_assignments(self):

    self.peak_list.clear()

    spectrum = self.settings.spectrum
    if spectrum == None:
      return
    
    atoms = sputil.hydrogen_atoms(spectrum.molecule)

    missing = 0
    self.stoppable_loop('atoms', 1)
    for atom in atoms:
      self.check_for_stop()
      missing = missing + self.show_missing_atom_assignments(atom, atoms)

    self.peak_list.heading['text'] = ('%d missing peaks\n' % missing
				      + '  ' + self.line_format %
				      ('assignment', 'dist', 'S/N',
				      '   frequency       '))

  # ---------------------------------------------------------------------------
  #
  def show_missing_atom_assignments(self, atom, atoms):

    spectrum = self.settings.spectrum
    if spectrum == None:
      return
    
    max_distance = self.settings.max_distance
    missing = 0
    close_list = close_atoms(atom, atoms, self.model, max_distance)
    for close_atom in close_list:
	if close_atom != atom:
	  assignment = self.missing_assignment(atom, close_atom, spectrum)
	  if assignment and not spectrum.find_peak(assignment):
	    line = self.missing_peak_format(assignment)
	    self.peak_list.append(line, assignment)
	    missing = missing + 1
    return missing

  # ---------------------------------------------------------------------------
  #
  def missing_assignment(self, atom1, atom2, spectrum):

    hres = noesy.atom_assignment(spectrum.condition, (atom1, atom2))
    if self.settings.noesy_hsqc:
      return noesy.extend_noesy_assignment(spectrum, hres)
    return hres
      
  # ---------------------------------------------------------------------------
  #
  def show_selected_peak_assignments(self):

    if self.settings.spectrum == None:
      return
    
    self.stoppable_call(self.show_peaks,
			self.show_alternative_assignments, 'selected peaks',
			self.settings.spectrum.selected_peaks())

  # ---------------------------------------------------------------------------
  #
  def show_alternative_assignments(self, peak, min_assignments = 0):

    alist = self.possible_assignments(peak)
    if len(alist) < min_assignments:
      return

    alist2 = []
    spectrum = peak.spectrum
    for a in alist:
      if not spectrum.find_peak(a):
	alist2.append(a)

    if len(alist2) + 1 >= min_assignments:
      self.peak_list.append(self.peak_format(peak), peak)
      for a in alist2:
	line = self.alternate_assignment_format(a, peak)
	self.peak_list.append(line, a)
      return 1
    return 0

  # ---------------------------------------------------------------------------
  #
  def possible_assignments(self, peak):
    if self.settings.noesy_hsqc:
      close_assignments = noesy.close_noesy_hsqc_assignments
    else:
      close_assignments = noesy.close_noesy_assignments
    return close_assignments(peak.spectrum, peak.frequency,
			     self.settings.ppm_range, self.model,
			     self.settings.max_distance)

  # ---------------------------------------------------------------------------
  #
  def peak_format(self, peak):

    name = peak.assignment
    resonances  = peak.resonances()
    dist = self.assignment_distance(resonances)
    distance = pyutil.number_string('%.2f', dist)
    if peak.data_height == None:
      strength = '%7s' % ''
    else:
      strength = '%7.1f' % (peak.data_height / peak.spectrum.noise)
    offset = sputil.resonance_offset_string(peak.frequency, resonances)

    return '  ' + self.line_format % (name, distance, strength, offset)

  # ---------------------------------------------------------------------------
  #
  def alternate_assignment_format(self, assignment, peak):

    name = sputil.assignment_name(assignment)
    dist = self.assignment_distance(assignment)
    distance = pyutil.number_string('%.2f', dist)
    strength = ''
    offset = sputil.resonance_offset_string(peak.frequency, assignment)

    return 'a ' + self.line_format % (name, distance, strength, offset)

  # ---------------------------------------------------------------------------
  #
  def missing_peak_format(self, assignment):

    if self.settings.spectrum == None:
      return ''

    name = sputil.assignment_name(assignment)
    dist = self.assignment_distance(assignment)
    distance = pyutil.number_string('%.2f', dist)
    spectrum = self.settings.spectrum
    freq = sputil.assignment_position(assignment, spectrum)
    strength = '%7.1f' % (spectrum.data_height(freq) / spectrum.noise)
    frequency = pyutil.sequence_string(freq, ' %7.3f')

    return 'm ' + self.line_format % (name, distance, strength, frequency)

  # ---------------------------------------------------------------------------
  #
  def assignment_distance(self, assignment):

    if self.settings.spectrum == None:
      return None

    if self.settings.noesy_hsqc:
      a1, aCN, a2 = noesy.noesy_hsqc_axes(self.settings.spectrum)
    else:
      a1 = 0
      a2 = 1
    r1 = assignment[a1]
    r2 = assignment[a2]
    if r1 and r2:
      return self.model.distance((r1.group.number, r1.atom.name),
                                 (r2.group.number, r2.atom.name))
    return None

# -----------------------------------------------------------------------------
#
def close_atoms(atom, atoms, pdb_model, max_distance):

  num_atom_1 = (atom.group.number, atom.name)
  atom_list = []
  for a in atoms:
    num_atom_2 = (a.group.number, a.name)
    d = pdb_model.distance(num_atom_1, num_atom_2)
    if d != None and d <= max_distance:
      atom_list.append(a)

  return atom_list

# -----------------------------------------------------------------------------
#
def show_dialog():
  tkutil.the_dialog(assignment_distance_dialog).show_window(1)
